@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "pdf-annot-imp.h"
@@ -452,10 +452,22 @@ void pdf_filter_annot_contents(fz_context *ctx, pdf_document *doc, pdf_annot *an
 	}
 }
 
+/* REDACTIONS */
+
+struct redact_filter_state {
+	pdf_filter_options filter_opts;
+	pdf_sanitize_filter_options sanitize_opts;
+	pdf_filter_factory filter_list[2];
+	pdf_page *page;
+	pdf_annot *target; // NULL if all
+	int line_art;
+};
+
 static void
 pdf_redact_end_page(fz_context *ctx, fz_buffer *buf, void *opaque)
 {
-	pdf_page *page = opaque;
+	struct redact_filter_state *red = opaque;
+	pdf_page *page = red->page;
 	pdf_annot *annot;
 	pdf_obj *qp;
 	int i, n;
@@ -464,6 +476,8 @@ pdf_redact_end_page(fz_context *ctx, fz_buffer *buf, void *opaque)
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 	{
+		if (red->target != NULL && red->target != annot)
+			continue;
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 		{
 			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
@@ -496,7 +510,8 @@ pdf_redact_end_page(fz_context *ctx, fz_buffer *buf, void *opaque)
 static int
 pdf_redact_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, fz_matrix trm, fz_matrix ctm, fz_rect bbox)
 {
-	pdf_page *page = opaque;
+	struct redact_filter_state *red = opaque;
+	pdf_page *page = red->page;
 	pdf_annot *annot;
 	pdf_obj *qp;
 	fz_rect r;
@@ -517,6 +532,8 @@ pdf_redact_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, f
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 	{
+		if (red->target != NULL && red->target != annot)
+			continue;
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 		{
 			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
@@ -644,10 +661,11 @@ pdf_redact_image_imp(fz_context *ctx, fz_matrix ctm, fz_image *image, fz_pixmap
 }
 
 static fz_image *
-pdf_redact_image_filter_remove(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image)
+pdf_redact_image_filter_remove(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image, fz_rect clip)
 {
 	fz_pixmap *redacted = NULL;
-	pdf_page *page = opaque;
+	struct redact_filter_state *red = opaque;
+	pdf_page *page = red->page;
 	pdf_annot *annot;
 	pdf_obj *qp;
 	fz_rect area;
@@ -660,6 +678,8 @@ pdf_redact_image_filter_remove(fz_context *ctx, void *opaque, fz_matrix ctm, con
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 	{
+		if (red->target != NULL && red->target != annot)
+			continue;
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 		{
 			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
@@ -688,11 +708,62 @@ pdf_redact_image_filter_remove(fz_context *ctx, void *opaque, fz_matrix ctm, con
 }
 
 static fz_image *
-pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image)
+pdf_redact_image_filter_remove_invisible(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image, fz_rect clip)
+{
+	fz_pixmap *redacted = NULL;
+	struct redact_filter_state *red = opaque;
+	pdf_page *page = red->page;
+	pdf_annot *annot;
+	pdf_obj *qp;
+	fz_rect area;
+	fz_rect r;
+	int i, n;
+
+	fz_var(redacted);
+
+	area = fz_transform_rect(fz_unit_rect, ctm);
+
+	/* Restrict the are of the image to that which can actually be seen. */
+	area = fz_intersect_rect(area, clip);
+
+	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
+	{
+		if (red->target != NULL && red->target != annot)
+			continue;
+		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
+		{
+			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
+			n = pdf_array_len(ctx, qp);
+			if (n > 0)
+			{
+				for (i = 0; i < n; i += 8)
+				{
+					r = fz_rect_from_quad(pdf_to_quad(ctx, qp, i));
+					r = fz_intersect_rect(r, area);
+					if (!fz_is_empty_rect(r))
+						return NULL;
+				}
+			}
+			else
+			{
+				r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
+				r = fz_intersect_rect(r, area);
+				if (!fz_is_empty_rect(r))
+					return NULL;
+			}
+		}
+	}
+
+	return fz_keep_image(ctx, image);
+}
+
+static fz_image *
+pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image, fz_rect clip)
 {
 	fz_pixmap *redacted = NULL;
 	fz_pixmap *mask = NULL;
-	pdf_page *page = opaque;
+	struct redact_filter_state *red = opaque;
+	pdf_page *page = red->page;
 	pdf_annot *annot;
 	pdf_obj *qp;
 	fz_quad area, q;
@@ -707,6 +778,8 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 	/* First see if we can redact the image completely */
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 	{
+		if (red->target != NULL && red->target != annot)
+			continue;
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 		{
 			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
@@ -735,6 +808,8 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 	{
 		for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 		{
+			if (red->target != NULL && red->target != annot)
+				continue;
 			if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 			{
 				qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
@@ -798,17 +873,24 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 	return fz_keep_image(ctx, image);
 }
 
+/* Returns 0 if area does not intersect with any of our redactions.
+ * Returns 2 if area is completely included within one of our redactions.
+ * Returns 1 otherwise. */
 static int
-rect_touches_redactions(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_rect area)
+rect_touches_redactions(fz_context *ctx, fz_rect area, struct redact_filter_state *red)
 {
 	pdf_annot *annot;
 	pdf_obj *qp;
 	fz_quad q;
-	fz_rect r;
+	fz_rect r, s;
 	int i, n;
+	pdf_page *page = red->page;
+	pdf_document *doc = red->page->doc;
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 	{
+		if (red->target != NULL && red->target != annot)
+			continue;
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 		{
 			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
@@ -819,17 +901,25 @@ rect_touches_redactions(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_r
 				{
 					q = pdf_to_quad(ctx, qp, i);
 					r = fz_rect_from_quad(q);
-					r = fz_intersect_rect(r, area);
-					if (!fz_is_empty_rect(r))
+					s = fz_intersect_rect(r, area);
+					if (!fz_is_empty_rect(s))
+					{
+						if (fz_contains_rect(r, area))
+							return 2;
 						return 1;
+					}
 				}
 			}
 			else
 			{
 				r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
-				r = fz_intersect_rect(r, area);
-				if (!fz_is_empty_rect(r))
+				s = fz_intersect_rect(r, area);
+				if (!fz_is_empty_rect(s))
+				{
+					if (fz_contains_rect(r, area))
+						return 2;
 					return 1;
+				}
 			}
 		}
 	}
@@ -837,14 +927,14 @@ rect_touches_redactions(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_r
 }
 
 static void
-pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page)
+pdf_redact_page_links(fz_context *ctx, struct redact_filter_state *red)
 {
 	pdf_obj *annots;
 	pdf_obj *link;
 	fz_rect area;
 	int k;
 
-	annots = pdf_dict_get(ctx, page->obj, PDF_NAME(Annots));
+	annots = pdf_dict_get(ctx, red->page->obj, PDF_NAME(Annots));
 	k = 0;
 	while (k < pdf_array_len(ctx, annots))
 	{
@@ -852,7 +942,7 @@ pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page)
 		if (pdf_dict_get(ctx, link, PDF_NAME(Subtype)) == PDF_NAME(Link))
 		{
 			area = pdf_dict_get_rect(ctx, link, PDF_NAME(Rect));
-			if (rect_touches_redactions(ctx, doc, page, area))
+			if (rect_touches_redactions(ctx, area, red))
 			{
 				pdf_array_delete(ctx, annots, k);
 				continue;
@@ -863,96 +953,150 @@ pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page)
 }
 
 static void
-pdf_redact_page_annotations(fz_context *ctx, pdf_document *doc, pdf_page *page)
+pdf_redact_page_annotations(fz_context *ctx, struct redact_filter_state *red)
 {
 	pdf_annot *annot;
 	fz_rect area;
 
 restart:
-	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
+	for (annot = pdf_first_annot(ctx, red->page); annot; annot = pdf_next_annot(ctx, annot))
 	{
 		if (pdf_annot_type(ctx, annot) == PDF_ANNOT_FREE_TEXT)
 		{
 			area = pdf_dict_get_rect(ctx, pdf_annot_obj(ctx, annot), PDF_NAME(Rect));
-			if (rect_touches_redactions(ctx, doc, page, area))
+			if (rect_touches_redactions(ctx, area, red))
 			{
-				pdf_delete_annot(ctx, page, annot);
+				pdf_delete_annot(ctx, red->page, annot);
 				goto restart;
 			}
 		}
 	}
 }
 
-int
-pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_options *opts)
+static int culler(fz_context *ctx, void *opaque, fz_rect bbox, fz_cull_type type)
+{
+	struct redact_filter_state *red = opaque;
+
+	switch (type)
+	{
+	case FZ_CULL_PATH_FILL:
+	case FZ_CULL_PATH_STROKE:
+	case FZ_CULL_PATH_FILL_STROKE:
+		if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_COVERED)
+			return (rect_touches_redactions(ctx, bbox, red) == 2);
+		else if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_TOUCHED)
+			return (rect_touches_redactions(ctx, bbox, red) != 0);
+		return 0;
+	default:
+		return 0;
+	}
+}
+
+static
+void init_redact_filter(fz_context *ctx, pdf_redact_options *redact_opts, struct redact_filter_state *red, pdf_page *page, pdf_annot *target)
+{
+	int black_boxes = redact_opts ? redact_opts->black_boxes : 0;
+	int image_method = redact_opts ? redact_opts->image_method : PDF_REDACT_IMAGE_PIXELS;
+	int line_art = redact_opts ? redact_opts->line_art : PDF_REDACT_LINE_ART_NONE;
+
+	memset(&red->filter_opts, 0, sizeof red->filter_opts);
+	memset(&red->sanitize_opts, 0, sizeof red->sanitize_opts);
+
+	red->filter_opts.recurse = 0; /* don't redact patterns, softmasks, and type3 fonts */
+	red->filter_opts.instance_forms = 1; /* redact xobjects with instancing */
+	red->filter_opts.ascii = 1;
+	red->filter_opts.opaque = red;
+	red->filter_opts.filters = red->filter_list;
+	if (black_boxes)
+		red->filter_opts.complete = pdf_redact_end_page;
+	red->line_art = line_art;
+
+	red->sanitize_opts.opaque = red;
+	red->sanitize_opts.text_filter = pdf_redact_text_filter;
+	if (image_method == PDF_REDACT_IMAGE_PIXELS)
+		red->sanitize_opts.image_filter = pdf_redact_image_filter_pixels;
+	if (image_method == PDF_REDACT_IMAGE_REMOVE)
+		red->sanitize_opts.image_filter = pdf_redact_image_filter_remove;
+	if (image_method == PDF_REDACT_IMAGE_REMOVE_UNLESS_INVISIBLE)
+		red->sanitize_opts.image_filter = pdf_redact_image_filter_remove_invisible;
+	red->sanitize_opts.culler = culler;
+
+	red->filter_list[0].filter = pdf_new_sanitize_filter;
+	red->filter_list[0].options = &red->sanitize_opts;
+	red->filter_list[1].filter = NULL;
+	red->filter_list[1].options = NULL;
+
+	red->page = page;
+	red->target = NULL;
+}
+
+static int
+pdf_apply_redaction_imp(fz_context *ctx, pdf_page *page, pdf_annot *target, pdf_redact_options *redact_opts)
 {
 	pdf_annot *annot;
 	int has_redactions = 0;
-	int black_boxes = 0;
-	int image_method = PDF_REDACT_IMAGE_PIXELS;
+	struct redact_filter_state red;
+	pdf_document *doc = page->doc;
 
-	pdf_filter_options options = { 0 };
-	pdf_sanitize_filter_options sopts = { 0 };
-	pdf_filter_factory list[2] = { 0 };
-
-	if (opts)
-	{
-		black_boxes = opts->black_boxes;
-		image_method = opts->image_method;
-	}
-
-	options.recurse = 0; /* don't redact patterns, softmasks, and type3 fonts */
-	options.instance_forms = 1; /* redact xobjects with instancing */
-	options.ascii = 1;
-	options.opaque = page;
-	options.filters = list;
-	if (black_boxes)
-		options.complete = pdf_redact_end_page;
-
-	list[0].filter = pdf_new_sanitize_filter;
-	list[0].options = &sopts;
-
-	sopts.opaque = page;
-	sopts.text_filter = pdf_redact_text_filter;
-	if (image_method == PDF_REDACT_IMAGE_PIXELS)
-		sopts.image_filter = pdf_redact_image_filter_pixels;
-	if (image_method == PDF_REDACT_IMAGE_REMOVE)
-		sopts.image_filter = pdf_redact_image_filter_remove;
-
-	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
+	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot)) {
+		if (target != NULL && target != annot)
+			continue;
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 			has_redactions = 1;
+	}
 
 	if (!has_redactions)
 		return 0;
 
-	pdf_begin_operation(ctx, doc, "Redact page");
+	init_redact_filter(ctx, redact_opts, &red, page, target);
+
+	if (target)
+		pdf_begin_operation(ctx, doc, "Apply redaction");
+	else
+		pdf_begin_operation(ctx, doc, "Apply redactions on page");
 	fz_try(ctx)
 	{
-		pdf_filter_page_contents(ctx, doc, page, &options);
-		pdf_redact_page_links(ctx, doc, page);
-		pdf_redact_page_annotations(ctx, doc, page);
+		pdf_filter_page_contents(ctx, doc, page, &red.filter_opts);
+		pdf_redact_page_links(ctx, &red);
+		pdf_redact_page_annotations(ctx, &red);
 
 		annot = pdf_first_annot(ctx, page);
 		while (annot)
 		{
-			if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
+			if (target == NULL || annot == target)
 			{
-				pdf_delete_annot(ctx, page, annot);
-				annot = pdf_first_annot(ctx, page);
-			}
-			else
-			{
-				annot = pdf_next_annot(ctx, annot);
+				if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
+				{
+					pdf_delete_annot(ctx, page, annot);
+					annot = pdf_first_annot(ctx, page);
+					continue;
+				}
 			}
+			annot = pdf_next_annot(ctx, annot);
 		}
 
 		doc->redacted = 1;
-	}
-	fz_always(ctx)
 		pdf_end_operation(ctx, doc);
+	}
 	fz_catch(ctx)
+	{
+		pdf_abandon_operation(ctx, doc);
 		fz_rethrow(ctx);
+	}
 
 	return 1;
 }
+
+int
+pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_options *redact_opts)
+{
+	if (page == NULL || page->doc != doc)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Can't redact a page not from the doc");
+	return pdf_apply_redaction_imp(ctx, page, NULL, redact_opts);
+}
+
+int
+pdf_apply_redaction(fz_context *ctx, pdf_annot *annot, pdf_redact_options *redact_opts)
+{
+	return pdf_apply_redaction_imp(ctx, annot->page, annot, redact_opts);
+}
