@@ -69,7 +69,7 @@ typedef struct filter_gstate
 {
 	struct filter_gstate *next;
 	int pushed;
-	int empty_clip_region;
+	fz_rect clip_rect;
 	pdf_filter_gstate pending;
 	pdf_filter_gstate sent;
 } filter_gstate;
@@ -266,7 +266,7 @@ static void filter_flush(fz_context *ctx, pdf_sanitize_processor *p, int flush)
 	int i;
 
 	/* No point in sending anything if we're clipping it away! */
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (gstate->pushed == 0)
@@ -874,7 +874,8 @@ push_adjustment_to_array(fz_context *ctx, pdf_sanitize_processor *p, pdf_obj *ar
 {
 	if (p->Tm_adjust == 0)
 		return;
-	pdf_array_push_real(ctx, arr, p->Tm_adjust * 1000);
+	/* Negate the number here, because it will be subtracted upon use. */
+	pdf_array_push_real(ctx, arr, -p->Tm_adjust * 1000);
 	p->Tm_adjust = 0;
 }
 
@@ -905,7 +906,7 @@ filter_show_string(fz_context *ctx, pdf_sanitize_processor *p, unsigned char *bu
 		}
 		if (i != len)
 		{
-			adjust_text(ctx, p, p->tos.char_tx, p->tos.char_ty);
+			adjust_text(ctx, p, p->tos.char_tx / p->gstate->pending.text.scale, p->tos.char_ty / p->gstate->pending.text.scale);
 			i += inc;
 		}
 		if (removed_space)
@@ -962,7 +963,7 @@ filter_show_text(fz_context *ctx, pdf_sanitize_processor *p, pdf_obj *text)
 					}
 					if (j != len)
 					{
-						adjust_text(ctx, p, p->tos.char_tx, p->tos.char_ty);
+						adjust_text(ctx, p, -p->tos.char_tx / p->gstate->pending.text.scale, -p->tos.char_ty / p->gstate->pending.text.scale);
 						j += inc;
 					}
 					if (removed_space)
@@ -971,7 +972,7 @@ filter_show_text(fz_context *ctx, pdf_sanitize_processor *p, pdf_obj *text)
 			}
 			else
 			{
-				float tadj = - pdf_to_real(ctx, item) * gstate->pending.text.size * 0.001f;
+				float tadj = pdf_to_real(ctx, item) * gstate->pending.text.size * 0.001f / p->gstate->pending.text.scale;
 				if (fontdesc->wmode == 0)
 				{
 					adjust_text(ctx, p, tadj, 0);
@@ -1001,7 +1002,7 @@ pdf_filter_w(fz_context *ctx, pdf_processor *proc, float linewidth)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.linewidth = linewidth;
@@ -1013,7 +1014,7 @@ pdf_filter_j(fz_context *ctx, pdf_processor *proc, int linejoin)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.linejoin = linejoin;
@@ -1025,7 +1026,7 @@ pdf_filter_J(fz_context *ctx, pdf_processor *proc, int linecap)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.linecap = linecap;
@@ -1037,7 +1038,7 @@ pdf_filter_M(fz_context *ctx, pdf_processor *proc, float miterlimit)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.miterlimit = miterlimit;
@@ -1048,7 +1049,7 @@ pdf_filter_d(fz_context *ctx, pdf_processor *proc, pdf_obj *array, float phase)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1061,7 +1062,7 @@ pdf_filter_ri(fz_context *ctx, pdf_processor *proc, const char *intent)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1074,7 +1075,7 @@ pdf_filter_gs_OP(fz_context *ctx, pdf_processor *proc, int b)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1087,7 +1088,7 @@ pdf_filter_gs_op(fz_context *ctx, pdf_processor *proc, int b)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1100,7 +1101,7 @@ pdf_filter_gs_OPM(fz_context *ctx, pdf_processor *proc, int i)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1113,7 +1114,7 @@ pdf_filter_gs_UseBlackPtComp(fz_context *ctx, pdf_processor *proc, pdf_obj *name
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1126,7 +1127,7 @@ pdf_filter_i(fz_context *ctx, pdf_processor *proc, float flatness)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1139,7 +1140,7 @@ pdf_filter_gs_begin(fz_context *ctx, pdf_processor *proc, const char *name, pdf_
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, FLUSH_ALL);
@@ -1153,7 +1154,7 @@ pdf_filter_gs_BM(fz_context *ctx, pdf_processor *proc, const char *blendmode)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_BM)
@@ -1165,7 +1166,7 @@ pdf_filter_gs_CA(fz_context *ctx, pdf_processor *proc, float alpha)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_CA)
@@ -1177,7 +1178,7 @@ pdf_filter_gs_ca(fz_context *ctx, pdf_processor *proc, float alpha)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_ca)
@@ -1189,7 +1190,7 @@ pdf_filter_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smask, float
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_SMask)
@@ -1201,7 +1202,7 @@ pdf_filter_gs_end(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_end)
@@ -1215,7 +1216,7 @@ pdf_filter_q(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_push(ctx, p);
@@ -1228,7 +1229,7 @@ pdf_filter_cm(fz_context *ctx, pdf_processor *proc, float a, float b, float c, f
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	fz_matrix ctm;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	/* If we're being given an identity matrix, don't bother sending it */
@@ -1252,14 +1253,12 @@ pdf_filter_m(fz_context *ctx, pdf_processor *proc, float x, float y)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_moveto(ctx, p->path, x, y);
 	if (p->options->culler)
-	{
-		fz_moveto(ctx, p->path, x, y);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_m)
@@ -1271,14 +1270,12 @@ pdf_filter_l(fz_context *ctx, pdf_processor *proc, float x, float y)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_lineto(ctx, p->path, x, y);
 	if (p->options->culler)
-	{
-		fz_lineto(ctx, p->path, x, y);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_l)
@@ -1290,14 +1287,12 @@ pdf_filter_c(fz_context *ctx, pdf_processor *proc, float x1, float y1, float x2,
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_curveto(ctx, p->path, x1, y1, x2, y2, x3, y3);
 	if (p->options->culler)
-	{
-		fz_curveto(ctx, p->path, x1, y1, x2, y2, x3, y3);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_c)
@@ -1309,14 +1304,12 @@ pdf_filter_v(fz_context *ctx, pdf_processor *proc, float x2, float y2, float x3,
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_curvetov(ctx, p->path, x2, y2, x3, y3);
 	if (p->options->culler)
-	{
-		fz_curvetov(ctx, p->path, x2, y2, x3, y3);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_v)
@@ -1328,14 +1321,12 @@ pdf_filter_y(fz_context *ctx, pdf_processor *proc, float x1, float y1, float x3,
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_curvetoy(ctx, p->path, x1, y1, x3, y3);
 	if (p->options->culler)
-	{
-		fz_curvetoy(ctx, p->path, x1, y1, x3, y3);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_y)
@@ -1347,14 +1338,12 @@ pdf_filter_h(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_closepath(ctx, p->path);
 	if (p->options->culler)
-	{
-		fz_closepath(ctx, p->path);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_h)
@@ -1366,14 +1355,12 @@ pdf_filter_re(fz_context *ctx, pdf_processor *proc, float x, float y, float w, f
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_rectto(ctx, p->path, x, y, x+w, y+h);
 	if (p->options->culler)
-	{
-		fz_rectto(ctx, p->path, x, y, x+w, y+h);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_re)
@@ -1441,7 +1428,7 @@ cull_replay_rectto(fz_context *ctx, void *arg, float x1, float y1, float x2, flo
 	pdf_sanitize_processor *p = (pdf_sanitize_processor *)arg;
 
 	if (p->chain->op_re)
-		p->chain->op_re(ctx, p->chain, x1, y1, x2, y2);
+		p->chain->op_re(ctx, p->chain, x1, y1, x2-x1, y2-y1);
 }
 
 typedef struct
@@ -1468,13 +1455,15 @@ end_segment(fz_context *ctx, segmenter_data_t *sd)
 		cull_replay_rectto
 	};
 	fz_rect r;
+	const fz_stroke_state *st;
 
 	if (sd->segment == NULL)
 		return;
 
-	r = fz_bound_path(ctx, sd->segment, (sd->type == FZ_CULL_PATH_STROKE || sd->type == FZ_CULL_PATH_FILL_STROKE) ? &sd->sstate : NULL, sd->ctm);
+	st = (sd->type == FZ_CULL_PATH_STROKE || sd->type == FZ_CULL_PATH_FILL_STROKE) ? &sd->sstate : NULL;
+	r = fz_bound_path(ctx, sd->segment, st, sd->ctm);
 
-	if (sd->p->options->culler(ctx, sd->p->options->opaque, r, sd->type))
+	if (sd->p->options->culler && sd->p->options->culler(ctx, sd->p->options->opaque, r, sd->type))
 	{
 		/* This segment can be skipped */
 	}
@@ -1565,7 +1554,12 @@ cull_path(fz_context *ctx, pdf_sanitize_processor *p, int type)
 	};
 
 	if (p->options->culler == NULL)
+	{
+		fz_drop_path(ctx, p->path);
+		p->path = NULL;
+		p->path = fz_new_path(ctx);
 		return 0;
+	}
 
 	sd.ctm = fz_concat(p->gstate->pending.ctm, p->gstate->sent.ctm);
 	sd.ctm = fz_concat(sd.ctm, p->transform);
@@ -1610,7 +1604,7 @@ pdf_filter_S(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_STROKE))
@@ -1626,7 +1620,7 @@ pdf_filter_s(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_STROKE))
@@ -1642,15 +1636,13 @@ pdf_filter_F(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL))
 		return;
 
 	filter_flush(ctx, p, FLUSH_FILL);
-	if (p->gstate->empty_clip_region)
-		return;
 	if (p->chain->op_F)
 		p->chain->op_F(ctx, p->chain);
 }
@@ -1660,7 +1652,7 @@ pdf_filter_f(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL))
@@ -1676,7 +1668,7 @@ pdf_filter_fstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL))
@@ -1692,7 +1684,7 @@ pdf_filter_B(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1708,7 +1700,7 @@ pdf_filter_Bstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1724,7 +1716,7 @@ pdf_filter_b(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1740,7 +1732,7 @@ pdf_filter_bstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1756,15 +1748,12 @@ pdf_filter_n(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
-	if (p->options->culler)
-	{
-		fz_drop_path(ctx, p->path);
-		p->path = NULL;
-		p->path = fz_new_path(ctx);
-	}
+	fz_drop_path(ctx, p->path);
+	p->path = NULL;
+	p->path = fz_new_path(ctx);
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_n)
@@ -1777,8 +1766,15 @@ static void
 pdf_filter_W(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
+	fz_rect r;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
+		return;
+
+	r = fz_bound_path(ctx, p->path, NULL, p->gstate->pending.ctm);
+	p->gstate->clip_rect = fz_intersect_rect(p->gstate->clip_rect, r);
+
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_CLIP_PATH))
@@ -1793,8 +1789,15 @@ static void
 pdf_filter_Wstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
+	fz_rect r;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
+		return;
+
+	r = fz_bound_path(ctx, p->path, NULL, p->gstate->pending.ctm);
+	p->gstate->clip_rect = fz_intersect_rect(p->gstate->clip_rect, r);
+
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_CLIP_PATH))
@@ -1812,7 +1815,7 @@ pdf_filter_BT(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1830,7 +1833,7 @@ pdf_filter_ET(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (!p->BT_pending)
@@ -1848,7 +1851,7 @@ pdf_filter_ET(fz_context *ctx, pdf_processor *proc)
 		 * clip path at all. We actually want it to be an empty clip path,
 		 * but we can't easily achieve that. So we have to do the clipping
 		 * ourselves. */
-		p->gstate->empty_clip_region = 1;
+		p->gstate->clip_rect = fz_empty_rect;
 	}
 	p->BT_pending = 0;
 	if (p->options->after_text_object)
@@ -1882,7 +1885,7 @@ pdf_filter_Tc(fz_context *ctx, pdf_processor *proc, float charspace)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1894,7 +1897,7 @@ pdf_filter_Tw(fz_context *ctx, pdf_processor *proc, float wordspace)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1908,7 +1911,7 @@ pdf_filter_Tz(fz_context *ctx, pdf_processor *proc, float scale)
 	 * in the gstate. */
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1920,7 +1923,7 @@ pdf_filter_TL(fz_context *ctx, pdf_processor *proc, float leading)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1932,7 +1935,7 @@ pdf_filter_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_font_d
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1950,7 +1953,7 @@ pdf_filter_Tr(fz_context *ctx, pdf_processor *proc, int render)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1962,7 +1965,7 @@ pdf_filter_Ts(fz_context *ctx, pdf_processor *proc, float rise)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1976,7 +1979,7 @@ pdf_filter_Td(fz_context *ctx, pdf_processor *proc, float tx, float ty)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	p->Tm_adjust = 0;
@@ -1995,7 +1998,7 @@ pdf_filter_TD(fz_context *ctx, pdf_processor *proc, float tx, float ty)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	p->gstate->pending.text.leading = -ty;
@@ -2007,7 +2010,7 @@ pdf_filter_Tm(fz_context *ctx, pdf_processor *proc, float a, float b, float c, f
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_tos_set_matrix(&p->tos, a, b, c, d, e, f);
@@ -2021,7 +2024,7 @@ pdf_filter_Tstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	p->Tm_adjust = 0;
@@ -2041,7 +2044,7 @@ pdf_filter_TJ(fz_context *ctx, pdf_processor *proc, pdf_obj *array)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_show_text(ctx, p, array);
@@ -2052,7 +2055,7 @@ pdf_filter_Tj(fz_context *ctx, pdf_processor *proc, char *str, size_t len)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_show_string(ctx, p, (unsigned char *)str, len);
@@ -2064,7 +2067,7 @@ pdf_filter_squote(fz_context *ctx, pdf_processor *proc, char *str, size_t len)
 	/* Note, we convert all T' operators to (maybe) a T* and a Tj */
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	/* need to flush text state and clear adjustment before we emit a newline */
@@ -2087,7 +2090,7 @@ pdf_filter_dquote(fz_context *ctx, pdf_processor *proc, float aw, float ac, char
 	 * (maybe) Tc, (maybe) Tw and a Tj. */
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	/* need to flush text state and clear adjustment before we emit a newline */
@@ -2112,7 +2115,7 @@ pdf_filter_d0(fz_context *ctx, pdf_processor *proc, float wx, float wy)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2125,7 +2128,7 @@ pdf_filter_d1(fz_context *ctx, pdf_processor *proc, float wx, float wy, float ll
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2181,7 +2184,7 @@ pdf_filter_CS(fz_context *ctx, pdf_processor *proc, const char *name, fz_colorsp
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.CS.name, name, sizeof gstate->pending.CS.name);
@@ -2196,7 +2199,7 @@ pdf_filter_cs(fz_context *ctx, pdf_processor *proc, const char *name, fz_colorsp
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.cs.name, name, sizeof gstate->pending.cs.name);
@@ -2212,7 +2215,7 @@ pdf_filter_SC_pattern(fz_context *ctx, pdf_processor *proc, const char *name, pd
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.SC.name, name, sizeof gstate->pending.SC.name);
@@ -2231,7 +2234,7 @@ pdf_filter_sc_pattern(fz_context *ctx, pdf_processor *proc, const char *name, pd
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.sc.name, name, sizeof gstate->pending.sc.name);
@@ -2249,7 +2252,7 @@ pdf_filter_SC_shade(fz_context *ctx, pdf_processor *proc, const char *name, fz_s
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.SC.name, name, sizeof gstate->pending.SC.name);
@@ -2265,7 +2268,7 @@ pdf_filter_sc_shade(fz_context *ctx, pdf_processor *proc, const char *name, fz_s
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.sc.name, name, sizeof gstate->pending.sc.name);
@@ -2282,7 +2285,7 @@ pdf_filter_SC_color(fz_context *ctx, pdf_processor *proc, int n, float *color)
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.SC.name[0] = 0;
@@ -2300,7 +2303,7 @@ pdf_filter_sc_color(fz_context *ctx, pdf_processor *proc, int n, float *color)
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.sc.name[0] = 0;
@@ -2317,7 +2320,7 @@ pdf_filter_G(fz_context *ctx, pdf_processor *proc, float g)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[1] = { g };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_CS(ctx, proc, "DeviceGray", fz_device_gray(ctx));
@@ -2330,7 +2333,7 @@ pdf_filter_g(fz_context *ctx, pdf_processor *proc, float g)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[1] = { g };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_cs(ctx, proc, "DeviceGray", fz_device_gray(ctx));
@@ -2343,7 +2346,7 @@ pdf_filter_RG(fz_context *ctx, pdf_processor *proc, float r, float g, float b)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[3] = { r, g, b };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_CS(ctx, proc, "DeviceRGB", fz_device_rgb(ctx));
@@ -2356,7 +2359,7 @@ pdf_filter_rg(fz_context *ctx, pdf_processor *proc, float r, float g, float b)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[3] = { r, g, b };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_cs(ctx, proc, "DeviceRGB", fz_device_rgb(ctx));
@@ -2369,7 +2372,7 @@ pdf_filter_K(fz_context *ctx, pdf_processor *proc, float c, float m, float y, fl
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[4] = { c, m, y, k };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_CS(ctx, proc, "DeviceCMYK", fz_device_cmyk(ctx));
@@ -2382,7 +2385,7 @@ pdf_filter_k(fz_context *ctx, pdf_processor *proc, float c, float m, float y, fl
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[4] = { c, m, y, k };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_cs(ctx, proc, "DeviceCMYK", fz_device_cmyk(ctx));
@@ -2396,7 +2399,7 @@ pdf_filter_BI(fz_context *ctx, pdf_processor *proc, fz_image *image, const char
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->options->culler)
@@ -2416,11 +2419,14 @@ pdf_filter_BI(fz_context *ctx, pdf_processor *proc, fz_image *image, const char
 		if (p->options->image_filter)
 		{
 			fz_matrix ctm = fz_concat(p->gstate->sent.ctm, p->transform);
-			image = p->options->image_filter(ctx, p->options->opaque, ctm, "<inline>", image);
+			image = p->options->image_filter(ctx, p->options->opaque, ctm, "<inline>", image, p->gstate->clip_rect);
 			if (image)
 			{
 				fz_try(ctx)
+				{
+					copy_resource(ctx, p, PDF_NAME(ColorSpace), colorspace);
 					p->chain->op_BI(ctx, p->chain, image, colorspace);
+				}
 				fz_always(ctx)
 					fz_drop_image(ctx, image);
 				fz_catch(ctx)
@@ -2429,6 +2435,7 @@ pdf_filter_BI(fz_context *ctx, pdf_processor *proc, fz_image *image, const char
 		}
 		else
 		{
+			copy_resource(ctx, p, PDF_NAME(ColorSpace), colorspace);
 			p->chain->op_BI(ctx, p->chain, image, colorspace);
 		}
 	}
@@ -2439,7 +2446,7 @@ pdf_filter_sh(fz_context *ctx, pdf_processor *proc, const char *name, fz_shade *
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->options->culler)
@@ -2466,7 +2473,7 @@ pdf_filter_Do_image(fz_context *ctx, pdf_processor *proc, const char *name, fz_i
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	fz_image *new_image;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->options->culler)
@@ -2486,7 +2493,7 @@ pdf_filter_Do_image(fz_context *ctx, pdf_processor *proc, const char *name, fz_i
 		if (p->options->image_filter)
 		{
 			fz_matrix ctm = fz_concat(p->gstate->sent.ctm, p->transform);
-			new_image = p->options->image_filter(ctx, p->options->opaque, ctm, name, image);
+			new_image = p->options->image_filter(ctx, p->options->opaque, ctm, name, image, p->gstate->clip_rect);
 		}
 		else
 		{
@@ -2539,7 +2546,7 @@ pdf_filter_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_o
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	fz_matrix transform;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, FLUSH_ALL);
@@ -2578,7 +2585,7 @@ pdf_filter_MP(fz_context *ctx, pdf_processor *proc, const char *tag)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2591,7 +2598,7 @@ pdf_filter_DP(fz_context *ctx, pdf_processor *proc, const char *tag, pdf_obj *ra
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2964,8 +2971,7 @@ pdf_new_sanitize_filter(
 
 	fz_try(ctx)
 	{
-		if (proc->options->culler)
-			proc->path = fz_new_path(ctx);
+		proc->path = fz_new_path(ctx);
 		proc->gstate = fz_malloc_struct(ctx, filter_gstate);
 		proc->gstate->pending.ctm = fz_identity;
 		proc->gstate->sent.ctm = fz_identity;
@@ -2977,6 +2983,7 @@ pdf_new_sanitize_filter(
 		proc->gstate->sent.text.size = -1;
 		proc->gstate->sent.stroke.linewidth = 1;
 		proc->gstate->sent.stroke.miterlimit = 10;
+		proc->gstate->clip_rect = fz_infinite_rect;
 	}
 	fz_catch(ctx)
 	{
